package translator.expression;

import java.util.ArrayList;

/**
 * Represents domains that consist of a set of sparse 
 * elements, for instance the domain (a,5,32,512*b). 
 * The sparse elements in the domain should be ordered
 * (even though we might not be able to test if the domain
 * is ordered in case there are unknown parameter values
 * in the domain, for instance)!
 * 
 * @author andrea
 *
 */

public class SparseExpressionRange implements ExpressionRange {

	Expression[] sparseElements;
	
	
	// ============ CONSTRUCTORS ==========================
	
	public SparseExpressionRange(Expression[] rangeElements) {
		this.sparseElements = rangeElements;
	}
	
	
	public SparseExpressionRange(ArrayList<Expression> elements) {
		
		this.sparseElements = new Expression[elements.size()];
		
		for(int i=this.sparseElements.length-1; i>=0; i--)
			this.sparseElements[i] = elements.remove(i);
		
	}
	
	//============= INHERTITED METHODS ======================
	
	public Domain copy() {
		Expression[] copiedRangeElements = new Expression[this.sparseElements.length];
		for(int i=0; i<this.sparseElements.length; i++)
			copiedRangeElements[i] = this.sparseElements[i].copy();
		
		return new SparseExpressionRange(copiedRangeElements);
	}

	public Domain evaluate() {
		boolean allElementsAreIntegers = true;
		
		for(int i=0; i<this.sparseElements.length; i++) {
			sparseElements[i] = this.sparseElements[i].evaluate();
			if(sparseElements[i].getType() != Expression.INT)
				allElementsAreIntegers = false;
		}
		if(allElementsAreIntegers) {
			int[] newIntRange = new int[this.sparseElements.length];
			for(int i=0; i<this.sparseElements.length; i++)
				newIntRange[i] = ((ArithmeticAtomExpression) this.sparseElements[i]).getConstant();
			return new SparseIntRange(newIntRange);
		}
		
		return this;
	}

	public Expression[] getLowerAndUpperBound() {
		return new Expression[] {this.sparseElements[0], this.sparseElements[this.sparseElements.length-1]};
	}
	
	public int getType() {
		return EXPR_SPARSE;
	}

	public String toString() {
		String s = "(";
		if(this.sparseElements.length > 0)
			s = ""+this.sparseElements[0].toString();
			
		for(int i=1;i<this.sparseElements.length; i++)
			s = s.concat(", "+sparseElements[i].toString());
		
		return s+")";
	}
	
	public boolean isConstantDomain() {
		return false;
	}
	
	public Domain insertValueForVariable(int value, String variableName) {
		
		boolean allSparseElementsAreInteger = true;
		
		for(int i=0; i<this.sparseElements.length; i++) {
			this.sparseElements[i] = this.sparseElements[i].insertValueForVariable(value, variableName);
			if(this.sparseElements[i].getType() != Expression.INT) {
				allSparseElementsAreInteger = false;
			}
		}
		
		if(allSparseElementsAreInteger) {
			int[] sparseIntElements = new int[this.sparseElements.length];
			for(int i=0; i<sparseIntElements.length; i++)
				sparseIntElements[i] = ((ArithmeticAtomExpression) this.sparseElements[i]).getConstant();
			
			return new SparseIntRange(sparseIntElements);
		}
		
		return this;
	}
	
	public Domain insertValueForVariable(boolean value, String variableName) {
		
		boolean allSparseElementsAreInteger = true;
		
		for(int i=0; i<this.sparseElements.length; i++) {
			this.sparseElements[i] = this.sparseElements[i].insertValueForVariable(value, variableName);
			if(this.sparseElements[i].getType() != Expression.INT) {
				allSparseElementsAreInteger = false;
			}
		}
		
		if(allSparseElementsAreInteger) {
			int[] sparseIntElements = new int[this.sparseElements.length];
			for(int i=0; i<sparseIntElements.length; i++)
				sparseIntElements[i] = ((ArithmeticAtomExpression) this.sparseElements[i]).getConstant();
			
			return new SparseIntRange(sparseIntElements);
		}
		
		return this;
	}
	
	
	public Domain replaceVariableWithDomain(String variableName, Domain newDomain) {
		return this;
	}
	
	public char isSmallerThanSameType(BasicDomain d) {
		
		SparseExpressionRange otherSparse = (SparseExpressionRange) d;
		
		if(this.sparseElements.length == otherSparse.sparseElements.length) {
			
			for(int i=0; i<this.sparseElements.length; i++) {
				
				if(this.sparseElements[i].getType() == otherSparse.sparseElements[i].getType()) {
					char difference = this.sparseElements[i].isSmallerThanSameType(otherSparse.sparseElements[i]);
					
					if(difference != Expression.EQUAL)
						return difference;
					
				}
				else return (this.sparseElements[i].getType() < otherSparse.sparseElements[i].getType()) ?
						Expression.SMALLER : Expression.BIGGER;
			}
			return Expression.EQUAL;
		}
		else return (this.sparseElements.length < otherSparse.sparseElements.length) ?
				Expression.SMALLER : Expression.BIGGER;
	}
	
}
